from dataclasses import dataclass, field
from datetime import date
from enum import Enum

import pytest

from serieux import deserialize
from serieux.ctx import Context
from serieux.features.clargs import CommandLineArguments
from serieux.features.tagged import Tagged

from ..definitions import Point


@dataclass
class Person:
    # Name of the person
    name: str

    # Age of the person
    age: int


@dataclass
class HeroProfile:
    beautiful: bool = True
    has_superpowers: bool = False
    loves_adventure: bool = True


class Material(str, Enum):
    WOOD = "wood"
    BRICK = "brick"
    CONCRETE = "concrete"
    STEEL = "steel"
    GLASS = "glass"


@dataclass
class House:
    # Date at which the house was built
    built: date

    # Building material
    material: Material


def test_simple():
    result = deserialize(
        Person,
        CommandLineArguments(["--name", "Jon", "--age", "27"]),
        Context(),
    )
    assert result == Person(name="Jon", age=27)


def test_help(capsys, file_regression):
    with pytest.raises(SystemExit):
        deserialize(Person, CommandLineArguments(["-h"]), Context())
    captured = capsys.readouterr()
    file_regression.check(captured.out + "\n=====\n" + captured.err)


def test_booleans():
    def f(*argv):
        return deserialize(HeroProfile, CommandLineArguments(argv), Context())

    assert f() == HeroProfile(True, False, True)
    assert f("--no-beautiful") == HeroProfile(False, False, True)
    assert f("--no-beautiful", "--has-superpowers", "--no-loves-adventure") == HeroProfile(
        False, True, False
    )


def test_misc_types():
    def f(*argv):
        return deserialize(House, CommandLineArguments(argv), Context())

    assert f("--built", "2025-01-01", "--material", "brick") == House(
        built=date(2025, 1, 1), material=Material.BRICK
    )
    with pytest.raises(SystemExit):
        f("--built", "2025-01-01", "--material", "invalid")


@dataclass
class Eat:
    """Stuffing your mouth."""

    food: str

    def do(self):
        return f"I eat {self.food}"


@dataclass
class Sleep:
    """Stuffing your brain with dreams."""

    hours: int

    def do(self):
        return f"I sleep {self.hours} hours"


@dataclass
class Act:
    """Do stuff!"""

    # What to do
    command: Tagged[Eat, "eat"] | Tagged[Sleep, "sleep"]  # noqa: F821

    # Do we do it fast?
    fast: bool = field(default=False, metadata={"argparse": {"alias": "-f"}})

    def text(self):
        return self.command.do() + (" fast" if self.fast else "")


@dataclass
class TalkingPoint(Point):
    def text(self):
        return f"Hi I am at ({self.x}, {self.y})"


def test_subcommands():
    def do(*args):
        result = deserialize(
            Tagged[Eat, "eat"] | Tagged[Sleep, "sleep"],
            CommandLineArguments(args),
            Context(),
        )
        return result.do()

    assert do("eat", "--food", "jam") == "I eat jam"
    assert do("sleep", "--hours", "8") == "I sleep 8 hours"


def test_args_plus_subcommands():
    def text(*args):
        result = deserialize(Act, CommandLineArguments(args), Context())
        return result.text()

    assert text("--fast", "eat", "--food", "jam") == "I eat jam fast"
    assert text("-f", "eat", "--food", "jam") == "I eat jam fast"
    assert text("sleep", "--hours", "8") == "I sleep 8 hours"


def test_subcommands_help(capsys, file_regression):
    with pytest.raises(SystemExit):
        deserialize(Act, CommandLineArguments(["-h"]), Context())
    captured = capsys.readouterr()
    file_regression.check(captured.out + "\n=====\n" + captured.err)


def test_subcommand_help(capsys, file_regression):
    with pytest.raises(SystemExit):
        deserialize(Act, CommandLineArguments(["eat", "-h"]), Context())
    captured = capsys.readouterr()
    file_regression.check(captured.out + "\n=====\n" + captured.err)


def test_sub_subcommands():
    def text(*args):
        result = deserialize(
            Tagged[Act, "act"] | Tagged[TalkingPoint, "point"],
            CommandLineArguments(args),
            Context(),
        )
        return result.text()

    assert text("act", "eat", "--food", "jam") == "I eat jam"
    assert text("act", "--fast", "eat", "--food", "jam") == "I eat jam fast"
    assert text("act", "sleep", "--hours", "8") == "I sleep 8 hours"
    assert text("point", "-x", "1", "-y", "2") == "Hi I am at (1, 2)"


@dataclass
class Word:
    word: str = field(metadata={"argparse": {"positional": True}})


def test_positional():
    result = deserialize(Word, CommandLineArguments(["amazing"]), Context())
    assert result == Word(word="amazing")


@dataclass
class Duck:
    quacks: int = field(metadata={"argparse": {"option": "-q"}})


def test_replace_option():
    result = deserialize(Duck, CommandLineArguments(["-q", "7"]), Context())
    assert result == Duck(quacks=7)



from dataclasses import field, fields, make_dataclass
from functools import reduce

from ovld import Medley, call_next, ovld, recurse

from ..ctx import Context
from ..exc import SerieuxError, ValidationError, ValidationExceptionGroup, merge_errors
from ..instructions import NewInstruction
from ..model import Modelizable, model
from ..utils import PRIO_HIGH, PRIO_LOW

#############
# Constants #
#############


Partial = NewInstruction["Partial"]


class NOT_GIVEN_T:
    pass


NOT_GIVEN = NOT_GIVEN_T()


class PartialBase:
    pass


class Sources:
    def __init__(self, *sources):
        self.sources = sources


@ovld
def partialize(t: type[Modelizable]):
    m = model(t)
    fields = [
        (
            f.name,
            partialize(f.type),
            field(default=NOT_GIVEN, metadata={"description": f.description}),
        )
        for f in m.fields
    ]
    fields.append(
        ("_serieux_ctx", Context, field(default=NOT_GIVEN, metadata={"serieux_metavar": "$ctx"}))
    )
    dc = make_dataclass(
        cls_name=f"Partial[{t.__name__}]",
        bases=(PartialBase,),
        fields=fields,
        namespace={"_constructor": m.constructor},
    )
    return dc


@ovld
def partialize(t: type[PartialBase]):  # pragma: no cover
    return t


@ovld
def partialize(t: object):
    return Partial[t]


###################
# Implementations #
###################


class PartialBuilding(Medley):
    @ovld(priority=PRIO_HIGH)
    def deserialize(self, t: type[Partial[object]], obj: object, ctx: Context, /):
        try:
            return call_next(t, obj, ctx)
        except SerieuxError as exc:
            return exc

    @ovld(priority=PRIO_HIGH)
    def deserialize(self, t: type[object], obj: Sources, ctx: Context, /):
        parts = [recurse(Partial[t], src, ctx) for src in obj.sources]
        rval = instantiate(reduce(merge, parts))
        if isinstance(rval, SerieuxError):
            raise rval
        return rval

    @ovld(priority=PRIO_LOW)
    def deserialize(self, t: type[PartialBase], obj: object, ctx: Context, /):
        # Fallback for alternative ways to deserialize the original object
        return recurse(t._constructor, obj, ctx)


@model.register
def _(p: type[Partial[object]]):
    return call_next(partialize(p.pushdown()))


######################
# Merge partial data #
######################


@ovld(priority=2)
def merge(x: object, y: SerieuxError):
    return y


@ovld(priority=2)
def merge(x: SerieuxError, y: object):
    return x


@ovld(priority=2)
def merge(x: SerieuxError, y: SerieuxError):
    return ValidationExceptionGroup("Some errors occurred", [x, y])


@ovld(priority=1)
def merge(x: object, y: NOT_GIVEN_T):
    return x


@ovld(priority=1)
def merge(x: NOT_GIVEN_T, y: object):
    return y


@ovld(priority=1)
def merge(x: NOT_GIVEN_T, y: NOT_GIVEN_T):
    return NOT_GIVEN


@ovld
def merge(x: PartialBase, y: PartialBase):
    if (xc := x._constructor) is not (yc := y._constructor):
        raise ValidationError(
            f"Cannot merge sources because of incompatible constructors: '{xc}', '{yc}'"
        )
    args = {}
    for f in fields(type(x)):
        xv = getattr(x, f.name)
        yv = getattr(y, f.name)
        args[f.name] = recurse(xv, yv)
    return type(x)(**args)


@ovld
def merge(x: PartialBase, y: object):
    if (xc := x._constructor) is not type(y):
        raise ValidationError(
            f"Cannot merge sources because of incompatible constructors: '{xc}', '{type(y)}'."
        )
    return recurse(x, type(x)(**vars(y)))


@ovld
def merge(x: object, y: PartialBase):
    if (yc := y._constructor) is not type(x):
        raise ValidationError(
            f"Cannot merge sources because of incompatible constructors: '{type(x)}', '{yc}'."
        )
    return recurse(type(y)(**vars(x)), y)


@ovld
def merge(x: dict, y: dict):
    result = dict(x)
    for k, v in y.items():
        result[k] = recurse(result.get(k, NOT_GIVEN), v)
    return result


@ovld
def merge(x: list, y: list):
    return x + y


@ovld
def merge(x: object, y: object):
    return y


############################
# Instantiate partial data #
############################


@ovld
def instantiate(xs: list):
    rval = []
    err = None
    for v in xs:
        value = recurse(v)
        if isinstance(value, SerieuxError):
            err = merge_errors(err, value)
        else:
            rval.append(value)
    return err if err else rval


@ovld
def instantiate(xs: dict):
    rval = {}
    err = None
    for k, v in xs.items():
        if v is NOT_GIVEN:
            continue
        value = recurse(v)
        if isinstance(value, SerieuxError):
            err = merge_errors(err, value)
        else:
            rval[k] = value
    return err if err else rval


@ovld
def instantiate(p: PartialBase):
    dc = p._constructor
    args = recurse({f.name: getattr(p, f.name) for f in fields(dc)})
    if isinstance(args, SerieuxError):
        return args
    try:
        return dc(**args)
    except Exception as exc:
        return ValidationError(exc=exc, ctx=p._serieux_ctx)


@ovld
def instantiate(x: object):
    return x


# Add as a default feature in serieux.Serieux
__default_features__ = PartialBuilding



import re
from dataclasses import field

from ovld import Medley, call_next, ovld, recurse
from ovld.dependent import Regexp

from ..ctx import AccessPath
from .lazy import LazyProxy


class Variables(AccessPath):
    refs: dict[tuple[str, ...], object] = field(default_factory=dict)

    def evaluate_reference(self, ref):
        def try_int(x):
            try:
                return int(x)
            except ValueError:
                return x

        stripped = ref.lstrip(".")
        dots = len(ref) - len(stripped)
        root = () if not dots else self.access_path[:-dots]
        parts = [try_int(x) for x in stripped.split(".")]
        return self.refs[(*root, *parts)]


class VariableInterpolation(Medley):
    @ovld(priority=3)
    def deserialize(self, typ: type[object], value: object, ctx: Variables):
        rval = call_next(typ, value, ctx)
        ctx.refs[ctx.access_path] = rval
        return rval

    @ovld(priority=2)
    def deserialize(self, typ: type[object], value: Regexp[r"^\$\{[^}]+\}$"], ctx: Variables):
        expr = value.lstrip("${").rstrip("}")

        def interpolate():
            value = ctx.evaluate_reference(expr)
            return recurse(typ, value, ctx)

        return LazyProxy(interpolate)

    @ovld(priority=1)
    def deserialize(self, typ: type[object], value: Regexp[r"\$\{[^}]+\}"], ctx: Variables):
        def interpolate():
            def repl(match):
                return str(ctx.evaluate_reference(match.group(1)))

            subbed = re.sub(r"\$\{([^}]+)\}", repl, value)
            return recurse(typ, subbed, ctx)

        return LazyProxy(interpolate)


# Add as a default feature in serieux.Serieux
__default_features__ = VariableInterpolation



import importlib

from ovld import Medley, call_next, ovld, recurse

from ..ctx import Context
from ..exc import ValidationError
from ..instructions import NewInstruction

#############
# Constants #
#############

TaggedSubclass = NewInstruction["TaggedSubclass"]


###################
# Implementations #
###################


def _resolve(ref, base, ctx):
    if ref is None:
        return base

    if (ncolon := ref.count(":")) == 0:
        mod_name = base.__module__
        symbol = ref
    elif ncolon == 1:
        mod_name, symbol = ref.split(":")
    else:
        raise ValidationError(f"Bad format for class reference: '{ref}'", ctx=ctx)
    try:
        mod = importlib.import_module(mod_name)
        return getattr(mod, symbol)
    except (ModuleNotFoundError, AttributeError) as exc:
        raise ValidationError(exc=exc, ctx=ctx)


class TaggedSubclassFeature(Medley):
    @ovld(priority=10)
    def serialize(self, t: type[TaggedSubclass], obj: object, ctx: Context, /):
        base = t.pushdown()
        if not isinstance(obj, base):
            raise ValidationError(f"'{obj}' is not a subclass of '{base}'", ctx=ctx)
        objt = type(obj)
        qn = objt.__qualname__
        if "." in qn:
            raise ValidationError("Only top-level symbols can be serialized", ctx=ctx)
        mod = objt.__module__
        rval = call_next(objt, obj, ctx)
        rval["class"] = f"{mod}:{qn}"
        return rval

    def deserialize(self, t: type[TaggedSubclass], obj: dict, ctx: Context, /):
        base = t.pushdown()
        obj = dict(obj)
        cls_name = obj.pop("class", None)
        actual_class = _resolve(cls_name, base, ctx)
        if not issubclass(actual_class, base):
            raise ValidationError(f"'{obj}' is not a subclass of '{base}'", ctx=ctx)
        return recurse(actual_class, obj, ctx)


# Add as a default feature in serieux.Serieux
__default_features__ = TaggedSubclassFeature



from dataclasses import dataclass
from datetime import date

import pytest

from serieux import Serieux
from serieux.features.interpol import VariableInterpolation, Variables

deserialize = (Serieux + VariableInterpolation)().deserialize


@dataclass
class Player:
    name: str
    nickname: str
    number: int


@dataclass
class Team:
    name: str
    rank: int
    forward: Player
    defender: Player
    goalie: Player


def test_simple_interpolate():
    data = {"name": "Robert", "nickname": "${name}", "number": 1}
    assert deserialize(Player, data, Variables()) == Player(
        name="Robert",
        nickname="Robert",
        number=1,
    )


def test_relative():
    data = {
        "name": "Team ${forward.nickname}",
        "rank": 7,
        "forward": {"name": "Igor", "nickname": "${.name}", "number": 1},
        "defender": {"name": "Robert", "nickname": "${.name}${.name}", "number": 2},
        "goalie": {"name": "Harold", "nickname": "Roldy", "number": "${..rank}"},
    }
    assert deserialize(Team, data, Variables()) == Team(
        name="Team Igor",
        rank=7,
        forward=Player(name="Igor", nickname="Igor", number=1),
        defender=Player(name="Robert", nickname="RobertRobert", number=2),
        goalie=Player(name="Harold", nickname="Roldy", number=7),
    )


def test_chain():
    data = [
        {"name": "Aaron", "nickname": "Ho", "number": 1},
        {"name": "Barbara", "nickname": "${0.nickname}s", "number": 2},
        {"name": "Cornelius", "nickname": "${1.nickname}s", "number": 3},
        {"name": "Dominic", "nickname": "${2.nickname}s", "number": 4},
    ]
    players = deserialize(list[Player], data, Variables())
    assert str(players[1].nickname) == "Hos"
    assert str(players[2].nickname) == "Hoss"
    assert str(players[3].nickname) == "Hosss"


def test_refer_to_object():
    data = [{"name": "Jon", "nickname": "Pork", "number": 1}, "${0}"]
    players = deserialize(list[Player], data, Variables())
    assert players[0] == players[1]


@dataclass
class DateMix:
    sdate: str
    ddate: date


def test_further_conversion():
    data = {"sdate": "2025-05-01", "ddate": "${sdate}"}
    dm = deserialize(DateMix, data, Variables())
    assert dm.ddate == date(2025, 5, 1)


def test_further_conversion_2():
    data = {"sdate": "2025-05", "ddate": "${sdate}-01"}
    dm = deserialize(DateMix, data, Variables())
    assert dm.ddate == date(2025, 5, 1)


def test_deadlock():
    data = {"name": "${nickname}", "nickname": "${name}", "number": 1}
    player = deserialize(Player, data, Variables())
    with pytest.raises(Exception, match="Deadlock"):
        player.name == "x"



from dataclasses import replace

from ovld.medley import ChainAll, KeepLast, Medley


class Context(Medley, default_combiner=KeepLast):
    follow = ChainAll()


class EmptyContext(Context):
    pass


class AccessPath(Context):
    full_path: tuple = ()

    @property
    def access_path(self):
        return tuple(k for _, k in self.full_path)

    def follow(self, objt, obj, field):
        return replace(self, full_path=(*self.full_path, (obj, field)))


empty = EmptyContext()



from dataclasses import dataclass

from ovld import Code, ovld, recurse

from .instructions import InstructionType
from .model import Modelizable, model


class Tell:
    def __lt__(self, other):
        return self.cost() < other.cost()

    def cost(self):
        return 1


@dataclass(frozen=True)
class TypeTell(Tell):
    t: type

    def gen(self, arg):
        return Code("isinstance($arg, $t)", arg=arg, t=self.t)


@dataclass(frozen=True)
class KeyTell(Tell):
    key: str

    def gen(self, arg):
        return Code("(isinstance($arg, dict) and $k in $arg)", arg=arg, k=self.key)

    def cost(self):
        return 2


@dataclass(frozen=True)
class KeyValueTell(Tell):
    key: str
    value: object

    def gen(self, arg):
        return Code(
            "(isinstance($arg, dict) and $k in $arg and $arg[$k] == $v)",
            arg=arg,
            k=self.key,
            v=self.value,
        )

    def cost(self):  # pragma: no cover
        return 3


@ovld
def tells(typ: type[int] | type[str] | type[bool] | type[float] | type[list] | type[dict]):
    return {TypeTell(typ)}


@ovld
def tells(typ: type[Modelizable]):
    m = model(typ)
    return {TypeTell(dict)} | {KeyTell(f.serialized_name) for f in m.fields}


@ovld(priority=-1)
def tells(m: type[InstructionType]):
    return recurse(m.pushdown())



from __future__ import annotations

from dataclasses import dataclass

from ovld import Dataclass, Dependent, ovld
from ovld.dependent import Regexp

from serieux import deserialize, schema, serialize
from serieux.model import Field, Model


@dataclass
class RGB:
    red: int
    green: int
    blue: int

    @classmethod
    def serieux_deserialize(cls, obj, ctx, call_next):
        if isinstance(obj, str):
            hex_str = obj.lstrip("#")
            red = int(hex_str[0:2], 16)
            green = int(hex_str[2:4], 16)
            blue = int(hex_str[4:6], 16)
            return RGB(red=red, green=green, blue=blue)
        else:
            return call_next(cls, obj, ctx)

    @classmethod
    def serieux_serialize(cls, obj, ctx, call_next):
        if 0 <= obj.red <= 255:
            return f"#{obj.red:02x}{obj.green:02x}{obj.blue:02x}"
        else:
            return call_next(cls, obj, ctx)

    @classmethod
    def serieux_schema(cls, ctx, call_next):
        return {
            "oneOf": [
                {"type": "string", "pattern": r"^#[0-9a-fA-F]{6}$"},
                call_next(cls, ctx),
            ]
        }


def test_custom_deserialize():
    assert deserialize(RGB, "#ff00ff") == RGB(red=255, green=0, blue=255)
    assert deserialize(RGB, {"red": 255, "green": 100, "blue": 100}) == RGB(
        red=255, green=100, blue=100
    )


def test_custom_serialize():
    assert serialize(RGB, RGB(red=255, green=0, blue=255)) == "#ff00ff"
    assert serialize(RGB, RGB(red=1000, green=0, blue=0)) == {"red": 1000, "green": 0, "blue": 0}


def test_custom_schema():
    assert schema(RGB).compile(root=False) == {
        "oneOf": [
            {"type": "string", "pattern": r"^#[0-9a-fA-F]{6}$"},
            {
                "type": "object",
                "properties": {
                    "red": {"type": "integer"},
                    "green": {"type": "integer"},
                    "blue": {"type": "integer"},
                },
                "required": ["red", "green", "blue"],
            },
        ]
    }


@dataclass
class RGBO:
    red: int
    green: int
    blue: int

    @classmethod
    @ovld
    def serieux_deserialize(cls, obj: Regexp[r"^#[0-9a-fA-F]{6}$"], ctx, call_next):
        hex_str = obj.lstrip("#")
        red = int(hex_str[0:2], 16)
        green = int(hex_str[2:4], 16)
        blue = int(hex_str[4:6], 16)
        return RGBO(red=red, green=green, blue=blue)

    @classmethod
    @ovld
    def serieux_serialize(
        cls, obj: Dependent[Dataclass, lambda rgb: 0 <= rgb.red <= 255], ctx, call_next
    ):
        assert 0 <= obj.red <= 255
        return f"#{obj.red:02x}{obj.green:02x}{obj.blue:02x}"


def test_custom_deserialize_o():
    assert deserialize(RGBO, "#ff00ff") == RGBO(red=255, green=0, blue=255)
    assert deserialize(RGBO, {"red": 255, "green": 100, "blue": 100}) == RGBO(
        red=255, green=100, blue=100
    )


def test_custom_serialize_o():
    assert serialize(RGBO, RGBO(red=255, green=0, blue=255)) == "#ff00ff"
    assert serialize(RGBO, RGBO(red=1000, green=0, blue=0)) == {"red": 1000, "green": 0, "blue": 0}


class RGBM:
    red: int
    green: int
    blue: int

    def __init__(self, red, green, blue):
        self.red = red
        self.green = green
        self.blue = blue

    @classmethod
    def serieux_model(cls, call_next):
        return Model.make(
            original_type=cls,
            fields=[
                Field(name="red", type=int, serialized_name="R"),
                Field(name="green", type=int, serialized_name="G"),
                Field(name="blue", type=int, serialized_name="B"),
            ],
            constructor=cls,
        )


def test_custom_deserialize_m():
    obj = deserialize(RGBM, {"R": 30, "G": 100, "B": 200})
    assert isinstance(obj, RGBM)
    assert obj.red == 30 and obj.green == 100 and obj.blue == 200



import importlib
from pathlib import Path
from typing import TYPE_CHECKING

from .auto import Auto
from .ctx import Context
from .exc import ValidationError, ValidationExceptionGroup
from .features.clargs import CommandLineArguments
from .features.lazy import DeepLazy, Lazy, LazyProxy
from .features.partial import Sources
from .features.tagged import Tagged
from .impl import BaseImplementation
from .instructions import InstructionType, NewInstruction
from .version import version as __version__


def _default_features():
    # Collect features declared in the __default_features__ global variables of the features
    # in features/. To see which ones this is on a cloned repo, you can run:
    # $ git grep __default_features__
    here = Path(__file__).parent
    features = []
    for name in (here / "features").glob("*.py"):
        if not name.stem.startswith("_"):
            mod = importlib.import_module(f"{__spec__.name}.features.{name.stem}")
            if feat := getattr(mod, "__default_features__", None):
                features.append(feat)
    return features


default_features = _default_features()


if TYPE_CHECKING:  # pragma: no cover
    from typing import TypeVar

    from .schema import Schema

    T = TypeVar("T")

    JSON = list["JSON"] | dict[str, "JSON"] | int | str | float | bool | None

    class _MC:
        def __add__(self, other) -> type["Serieux"]: ...

    class Serieux(metaclass=_MC):
        def dump(
            self, t: type[T], obj: object, ctx: Context = None, *, dest: Path = None
        ) -> JSON | None: ...

        def load(self, t: type[T], obj: object, ctx: Context = None) -> T: ...

        def serialize(self, t: type[T], obj: object, ctx: Context = None) -> JSON: ...

        def deserialize(self, t: type[T], obj: object, ctx: Context = None) -> T: ...

        def schema(self, t: type[T], ctx: Context = None) -> Schema[str, "JSON"]: ...

        def __add__(self, other) -> "Serieux": ...

else:

    class Serieux(BaseImplementation, *default_features):
        pass


serieux = Serieux()
serialize = serieux.serialize
deserialize = serieux.deserialize
schema = serieux.schema
load = serieux.load
dump = serieux.dump


__all__ = [
    "__version__",
    "Auto",
    "BaseImplementation",
    "CommandLineArguments",
    "Context",
    "DeepLazy",
    "deserialize",
    "InstructionType",
    "Lazy",
    "LazyProxy",
    "NewInstruction",
    "schema",
    "serialize",
    "serieux",
    "Serieux",
    "Sources",
    "Tagged",
    "ValidationError",
    "ValidationExceptionGroup",
]



import argparse
from dataclasses import MISSING, dataclass
from enum import Enum
from typing import get_args

from ovld import Medley, ovld, recurse

from ..ctx import Context
from ..exc import ValidationError
from ..instructions import strip_all
from ..model import Field, Modelizable, model
from ..utils import UnionAlias, clsstring
from .dotted import unflatten
from .tagged import Tagged

##################
# Implementation #
##################


def _compose(dest, new_part):
    return f"{dest}.{new_part}" if dest else new_part


@dataclass
class CommandLineArguments:
    arguments: list[str]


@ovld
def make_argument(t: type[int] | type[float] | type[str], partial: dict, model_field: Field):
    return {**partial, "type": t}


@ovld
def make_argument(t: type[bool], partial: dict, model_field: Field):
    partial.pop("metavar", None)
    partial.pop("type", None)
    return {**partial, "action": argparse.BooleanOptionalAction}


@ovld(priority=1)
def make_argument(t: type[Enum], partial: dict, model_field: Field):
    return {**partial, "type": str, "choices": [e.value for e in t]}


@ovld
def make_argument(t: type[object], partial: dict, model_field: Field):
    return {**partial, "type": str}


@ovld
def make_argument(t: type[UnionAlias], partial: dict, model_field: Field):
    return "subparser"


@ovld
def add_arguments(t: type[Modelizable], parser: argparse.ArgumentParser, dest: str):
    m = model(t)
    for field in m.fields:
        if field.name.startswith("_"):  # pragma: no cover
            continue

        name = field.name.replace("_", "-")
        typ = strip_all(field.type)
        meta = dict(field.metadata.get("argparse", {}))
        positional = meta.pop("positional", False)
        fdest = _compose(dest, field.name)
        fhelp = field.description or field.name
        mvar = name.split(".")[-1].upper()

        if positional:
            args = {"__args__": [fdest], "help": fhelp, "metavar": mvar}
        else:
            args = {
                "__args__": [f"--{name}" if len(name) > 1 else f"-{name}"],
                "help": fhelp,
                "metavar": mvar,
                "required": field.required,
                "dest": fdest,
            }

        if field.default is not MISSING:
            args["default"] = field.default

        args = make_argument(typ, args, field)
        if args == "subparser":
            add_arguments(typ, parser, _compose(dest, field.name))
        else:
            pos = args.pop("__args__")
            if opt := meta.pop("option", None):
                if not isinstance(opt, list):
                    opt = [opt]
                pos[:] = opt
            if alias := meta.pop("alias", None):
                if not isinstance(alias, list):
                    alias = [alias]
                pos.extend(alias)
            args.update(meta)
            parser.add_argument(*pos, **args)
    return parser


@ovld
def add_arguments(t: type[UnionAlias], parser: argparse.ArgumentParser, dest: str):
    options = get_args(t)
    if any(not issubclass(option, Tagged) for option in options):  # pragma: no cover
        raise ValidationError("All Union members must be Tagged to make a cli")

    subparsers = parser.add_subparsers(dest=_compose(dest, "class"))
    for opt in options:
        subparsers.required = True
        subparser = subparsers.add_parser(opt.tag, help=f"{opt.cls.__doc__ or opt.tag}")
        add_arguments(opt.cls, subparser, dest)


class FromArguments(Medley):
    @ovld(priority=1)
    def deserialize(self, t: type[object], obj: CommandLineArguments, ctx: Context):
        parser = argparse.ArgumentParser(description=t.__doc__ or f"Arguments for {clsstring(t)}")
        add_arguments(t, parser, "")
        ns = parser.parse_args(obj.arguments)
        values = {k: v for k, v in vars(ns).items() if v is not None}
        return recurse(t, unflatten(values), ctx)


# Add as a default feature in serieux.Serieux
__default_features__ = FromArguments
